import logbook
import sys
import os
from infi.pyutils.contexts import contextmanager
from infi.pyutils.decorators import wraps

from .compat import redirect_python_logging_to_logbook
from .processors import create_processor
from .handlers import RotatingFileHandler
from .formatters import create_default_formatter
from .plugins import _true
from .plugins.procname import get_procname

try:
    from .handlers import SyslogHandler
    _has_syslog_handler = True
except ImportError:
    _has_syslog_handler = False


def _not_time(s):
    return s != 'time'


def _syslog_pred(s):
    return s not in ['time', 'hostname', 'procname', 'process_id']


def create_rotating_file_handler(path, mode='a', encoding='utf-8', level=logbook.DEBUG, delay=False,
                                 max_size=1024 * 1024, backup_count=32, formatter_plugin_predicate=_true):
    """Convenience function to create a rotating file handler with the default formatter."""
    handler = RotatingFileHandler(filename=path, mode=mode, encoding=encoding, level=level, delay=delay,
                                  max_size=max_size, backup_count=backup_count, bubble=True)
    handler.formatter = create_default_formatter(formatter_plugin_predicate)
    return handler


def create_stream_handler(stream, level=logbook.INFO, formatter_plugin_predicate=_not_time):
    """Convenience function to create a stream handler with the default formatter (w/o time field)."""
    handler = logbook.StreamHandler(sys.stderr, level=level, bubble=True)
    handler.formatter = create_default_formatter(formatter_plugin_predicate)
    return handler


def create_stdout_handler(level=logbook.INFO, formatter_plugin_predicate=_not_time):
    """Convenience function to create an STDOUT handler with the default formatter (w/o time field)."""
    return create_stream_handler(sys.stdout, level, formatter_plugin_predicate)


def create_stderr_handler(level=logbook.INFO, formatter_plugin_predicate=_not_time):
    """Convenience function to create an STDERR handler with the default formatter (w/o time field)."""
    return create_stream_handler(sys.stderr, level, formatter_plugin_predicate)


def create_syslog_handler(facility=logbook.SyslogHandler.LOG_LOCAL1, buffer_size=1024, message_size=32768,
                          address=("127.0.0.1", 514), level=logbook.DEBUG, formatter_plugin_predicate=_syslog_pred):
    """Convenience function to create an syslog handler with the default formatter."""
    from socket import gethostname
    application_name = get_procname()
    if application_name is None:
        application_name = 'python'
    handler = SyslogHandler(facility=facility, host_name=gethostname(), application_name=application_name,
                            process_id=str(os.getpid()), address=address, level=level, bubble=True,
                            syslog_buffer_size=buffer_size, syslog_message_size=message_size)
    handler.formatter = create_default_formatter(formatter_plugin_predicate)
    return handler


@contextmanager
def script_logging_context(syslog=_has_syslog_handler, syslog_facility=logbook.SyslogHandler.LOG_LOCAL1,
                           syslog_buffer_size=1024, syslog_message_size=32768, syslog_address=("127.0.0.1", 514),
                           syslog_level=logbook.DEBUG, logfile=True, logfile_path="logfile", logfile_mode='a',
                           logfile_encoding='utf-8', logfile_level=logbook.DEBUG, logfile_delay=False,
                           logfile_max_size=1024 * 1024, logfile_backup_count=32, stderr=True,
                           stderr_level=logbook.INFO):
    """
    Context manager that creates a setup of logbook handlers based on the parameters received and sensible defaults.
    """
    from logbook.concurrency import enable_gevent
    enable_gevent()
    redirect_python_logging_to_logbook()

    processor = create_processor()
    flags = logbook.Flags(errors='silent')
    handlers = [logbook.NullHandler()]

    if syslog:
        handlers.append(create_syslog_handler(facility=syslog_facility, buffer_size=syslog_buffer_size,
                                              message_size=syslog_message_size, address=syslog_address,
                                              level=syslog_level))
    if logfile:
        handlers.append(create_rotating_file_handler(path=logfile_path, mode=logfile_mode,
                                                     encoding=logfile_encoding, level=logfile_level,
                                                     delay=logfile_delay, max_size=logfile_max_size,
                                                     backup_count=logfile_backup_count))
    if stderr:
        handlers.append(create_stderr_handler(level=stderr_level))

    with logbook.NestedSetup([processor, flags] + handlers).applicationbound():
        yield


def script_logging_decorator(func=None, *args, **kwargs):
    """Decorator equivalent for `script_logging_context`"""
    def decorate(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            with script_logging_context():
                return f(*args, **kwargs)
        return decorator
    if func is None:
        return decorate
    else:
        return decorate(func)



from infi.pyutils.decorators import wraps
from infi.pyutils.contexts import contextmanager
from .globals import get_time, get_logger


@contextmanager
def log_timing_context(title, logger=None, log_level="DEBUG"):
    """
    Log the execution time of the context. It generates two log messages: before yielding and after.
    :param title: str to add to the log messages
    :param logger: logger to use, or None if using the default logger
    :param log_level: str log level to use [TRACE|DEBUG|INFO|...]
    """
    global _logger
    if logger is None:
        logger = get_logger()
    log = getattr(logger, log_level.lower())
    t0 = get_time()
    log("started timing {} at {:.4f}".format(title, t0))
    try:
        yield
    finally:
        t1 = get_time()
        log("ended timing {} at {:.4f}. time taken: {:.4f} seconds".format(title, t1, t1 - t0))


def log_timing(func=None, logger=None, log_level="DEBUG"):
    """
    Decorator to log the decorated function execution time. See `log_timing_context` for more details.
    It can be used in several methods:
    ```
    @log_timing
    def foo():
        pass

    @log_timing(logger=my_logger)
    def foo():
        pass

    @log_timing(logger=my_logger, log_level='DEBUG')
    def foo():
        pass
    ```

    :param func: function to decorate
    :param logger: logger to use, or None if using the default logger
    :param log_level: str log level to use [TRACE|DEBUG|INFO|...]
    """
    def decorate(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            with log_timing_context("function {!r}".format(f.__name__), logger=logger, log_level=log_level):
                return f(*args, **kwargs)
        return wrapped
    if func is None:
        return decorate
    else:
        return decorate(func)



import pkg_resources

FORMATTER_PLUGIN_ENTRYPOINT = 'infi_logging_formatter_plugins'
INJECTOR_PLUGIN_ENTRYPOINT = 'infi_logging_injector_plugins'

_formatter_plugins = None
_injector_plugins = None


class InjectorPlugin(object):
    def inject(self, record):
        raise NotImplementedError()  # subclass must override this method


class FormatterPlugin(object):
    def get_value(self, record):
        raise NotImplementedError()  # subclass must override this method

    def get_format_string(self):
        raise NotImplementedError()  # subclass must override this method

    def get_format_key(self):
        raise NotImplementedError()  # subclass must override this method


def _true(name):
    return True


def load_plugins_by_entry_point(entry_point_key, predicate=_true, strict=False):
    plugins = dict()
    for entrypoint in pkg_resources.iter_entry_points(entry_point_key):
        if predicate(entrypoint.name):
            try:
                plugins[entrypoint.name] = entrypoint.load()
            except ImportError:
                if strict:
                    raise
    return plugins


def load_formatter_plugins(predicate=_true, strict=False):
    global _formatter_plugins
    if _formatter_plugins is None:
        _formatter_plugins = load_plugins_by_entry_point(FORMATTER_PLUGIN_ENTRYPOINT, predicate, strict)
    return _formatter_plugins


def load_injector_plugins(predicate=_true, strict=False):
    global _injector_plugins
    if _injector_plugins is None:
        _injector_plugins = load_plugins_by_entry_point(INJECTOR_PLUGIN_ENTRYPOINT, predicate, strict)
    return _injector_plugins


def get_injector_plugins(predicate=_true):
    plugins = load_injector_plugins()
    return dict((k, v) for k, v in plugins.items() if predicate(k))


def get_formatter_plugins(predicate=_true):
    plugins = load_formatter_plugins()
    return dict((k, v) for k, v in plugins.items() if predicate(k))


def clear_formatter_plugins():
    global _formatter_plugins
    _formatter_plugins = None


def clear_injector_plugins():
    global _injector_plugins
    _injector_plugins = None


def clear_plugins():
    clear_formatter_plugins()
    clear_injector_plugins()



from infi.logging.plugins import FormatterPlugin


class MessageFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        if record.formatted_exception:
            return record.message + "\n" + record.formatted_exception
        else:
            return record.message

    def get_format_string(self):
        return "{}"

    def get_format_key(self):
        return "message"



from infi.logging.plugins import InjectorPlugin, FormatterPlugin
try:
    from threading import get_ident
except ImportError:     # Python 2.x
    from thread import get_ident

THREAD_ID_KEY = 'thread_id'
THREAD_ID_MODULU = 32768


def get_thread_id_from_record(record):
    """
    :param record: logbook record
    :returns: current thread ID from record
    """
    return record.extra.get(THREAD_ID_KEY, record.thread)


def inject_thread_id(record):
    """
    Sets the current thread ID on `record.extra`.
    :param record: logbook record
    """
    record.extra[THREAD_ID_KEY] = get_ident()


class ThreadIDInjectorPlugin(InjectorPlugin):
    def inject(self, record):
        return inject_thread_id(record)


class ThreadIDFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return get_thread_id_from_record(record) % THREAD_ID_MODULU

    def get_format_string(self):
        return "{:0>5}"

    def get_format_key(self):
        return "thread_id"



from infi.logging.plugins import FormatterPlugin


def get_time_from_record(record):
    """
    :param record: logbook record
    :returns: time from record
    """
    return record.time


class TimeFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return get_time_from_record(record)

    def get_format_string(self):
        return "{:%Y-%m-%d %H:%M:%S}"

    def get_format_key(self):
        return "time"



import six
from logbook import Handler, StringFormatterHandlerMixin, NOTSET, DEBUG, INFO, NOTICE, WARNING, ERROR, CRITICAL
from ctypes import c_wchar_p, WinError, windll

from infi.registry import RegistryValueFactory, LocalComputer

RegisterEventSourceW = windll.advapi32.RegisterEventSourceW
DeregisterEventSource = windll.advapi32.DeregisterEventSource
ReportEventW = windll.advapi32.ReportEventW

# From http://msdn.microsoft.com/en-us/library/windows/desktop/aa363679%28v=VS.85%29.aspx
EVENTLOG_SUCCESS = 0x0000
EVENTLOG_AUDIT_FAILURE = 0x0010
EVENTLOG_AUDIT_SUCCESS = 0x0008
EVENTLOG_ERROR_TYPE = 0x0001
EVENTLOG_INFORMATION_TYPE = 0x0004
EVENTLOG_WARNING_TYPE = 0x0002


def register_application(app_name, message_file_path=None):
    """Registers an application as an event log source so it can later create events.

    :param app_name: Application name (use the same name later when registering the event source)
    :param message_file_path: resource DLL file path
    """
    # See http://msdn.microsoft.com/en-us/library/windows/desktop/aa363661%28v=VS.85%29.aspx

    if message_file_path is None:
        import sys
        from pkg_resources import resource_filename
        fname = "eventlog_amd64.dll" if sys.maxsize > (1 << 31) else "eventlog_x86.dll"
        message_file_path = resource_filename("infi.logging", fname)

    value_factory = RegistryValueFactory()
    event_log = LocalComputer().local_machine['SYSTEM']['CurrentControlSet']['Services']['Eventlog']
    event_log['Application'][app_name] = None
    app_key = event_log['Application'][app_name]
    app_key.values_store['EventMessageFile'] = value_factory.by_value(message_file_path)
    app_key.values_store['TypesSupported'] = value_factory.by_value(7)  # All severity types


def unregister_application(app_name):
    """Unregisters an application as a log source by removing it from the registry."""
    event_log = LocalComputer().local_machine['SYSTEM']['CurrentControlSet']['Services']['Eventlog']
    event_log['Application'].pop(app_name, None)


def register_event_source(app_name):
    """Registers an event source and returns an open handle. Raises WinError on error."""
    handle = RegisterEventSourceW(None, six.text_type(app_name))
    if handle == 0:
        raise WinError()
    return handle


def deregister_event_source(handle):
    """Closes an open event source handle."""
    DeregisterEventSource(handle)


def report_event(handle, type, category, event_id, strings, raw_data):
    """Reports an event to Windows event log. Raises WinError on error.

    :param handle: handle to an open event source, previously opened with register_event_source
    :param type: event type: one of the EVENTLOG_xxxx constants
    :param category: category to use (number). If using our resource DLL, leave this as 0.
    :param event_id: event ID to use (number).
    :param strings: a list of strings that will be formatted as the event message. IF using our resource DLL, only a single
                    string is supported.
    :param raw_data: a blob (string)
    """
    # See http://msdn.microsoft.com/en-us/library/windows/desktop/aa363679%28v=VS.85%29.aspx
    raw_data_len = 0 if raw_data is None else len(raw_data)
    strings_ptr = (c_wchar_p * len(strings))()
    for i in range(len(strings)):
        strings_ptr[i] = c_wchar_p(six.text_type(strings[i]))
    result = ReportEventW(handle, type, category, event_id, None, len(strings), raw_data_len, strings_ptr, raw_data)
    if result != 1:
        raise WinError()


class Win32EventLogHandler(Handler, StringFormatterHandlerMixin):
    default_format_string = "{record.message}"

    def __init__(self, application_name, level=NOTSET, format_string=None, filter=None, bubble=False):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)

        self._application_name = application_name
        self._default_type = EVENTLOG_INFORMATION_TYPE
        self._type_map = {
            DEBUG:      EVENTLOG_INFORMATION_TYPE,
            INFO:       EVENTLOG_INFORMATION_TYPE,
            NOTICE:     EVENTLOG_INFORMATION_TYPE,
            WARNING:    EVENTLOG_WARNING_TYPE,
            ERROR:      EVENTLOG_ERROR_TYPE,
            CRITICAL:   EVENTLOG_ERROR_TYPE
        }
        self._win32_handle = register_event_source(self._application_name)

    def close(self):
        if self._win32_handle != 0:
            deregister_event_source(self._win32_handle)
            self._win32_handle = 0

    def get_message_id(self, record):
        for dic in (record.extra, record.kwargs):
            id = dic.get('eventid', 0)
            if id != 0:
                return id

        # TODO: warn on debug that we have no id for this event.
        return 0

    def get_event_category(self, record):
        for dic in (record.extra, record.kwargs):
            category = dic.get('eventcategory', 0)
            if category != 0:
                return category
        return 0

    def get_event_type(self, record):
        return self._type_map.get(record.level, self._default_type)

    def get_raw_data(self, record):
        for dic in (record.extra, record.kwargs):
            raw_data = dic.get('raw_data', None)
            if raw_data is not None:
                return raw_data
        return None

    def emit(self, record):
        id = self.get_message_id(record)
        cat = self.get_event_category(record)
        type = self.get_event_type(record)
        message = self.format(record)
        raw_data = self.get_raw_data(record)
        report_event(self._win32_handle, type, cat, id, [message], raw_data)



import logbook
from infi.tracing import SyslogWriter


class SyslogHandler(logbook.Handler, logbook.StringFormatterHandlerMixin):
    """Handler that uses infi.tracing's `SyslogWriter` to truly asynchronously write to syslog."""

    def __init__(self, facility, host_name="", application_name="python", process_id="", address=("127.0.0.1", 514),
                 level=logbook.NOTSET, format_string=None, filter=None, bubble=False,
                 syslog_buffer_size=1024, syslog_message_size=32768):
        logbook.Handler.__init__(self, level, filter, bubble)
        logbook.StringFormatterHandlerMixin.__init__(self, format_string)

        self.writer = SyslogWriter(syslog_buffer_size, syslog_message_size, facility=facility, address=address,
                                   host_name=host_name, application_name=application_name, process_id=process_id,
                                   rfc5424=True)
        self.writer.start()

    def emit(self, record):
        self.writer.write(logbook.SyslogHandler.level_priority_map.get(record.level,
                                                                       logbook.SyslogHandler.LOG_WARNING),
                          self.format(record))

    def close(self):
        self.writer.stop()



"""
Slow greenlet logging utility
"""
import gevent
import greenlet
from functools import partial

from .globals import get_time, get_logger

_slow_greenlet_max_duration = 1.0
_last_switch_time = None


def _switch_time_tracer(logger, event, args):
    if event not in ('switch', 'throw'):
        # from greenlet docs: "For compatibility it is very important to unpack args tuple only when event is either
        # 'switch' or 'throw' and not when event is potentially something else.
        return
    (origin, target) = args
    global _slow_greenlet_max_duration, _last_switch_time
    now = get_time()
    if _last_switch_time and origin != gevent.get_hub():  # gevent.hub can block for as long as it wants
        duration = now - _last_switch_time
        if duration >= _slow_greenlet_max_duration:
            msg = "greenlet id {} was running for at least {:.4f} seconds"
            logger.warn(msg.format(id(origin), duration))
    _last_switch_time = now


def enable_slow_greenlet_log_warning(max_duration=1.0, logger=None):
    """
    Enables warnings about slow greenlet written to the log
    :param max_duration: maximum duration in seconds afterwhich a greenlet is considered slow
    :param logger: logger to use, or None if using the default logger
    """
    global _slow_greenlet_max_duration, _last_switch_time, _logger
    if logger is None:
        logger = get_logger()
    now = get_time()
    _slow_greenlet_max_duration = max_duration
    _last_switch_time = now
    current_id = id(gevent.getcurrent())
    logger.debug("enabling logging of greenlet switching, current greenlet (main) is {}".format(current_id))
    greenlet.settrace(partial(_switch_time_tracer, logger))


def disable_slow_greenlet_log_warning():
    """Disables the slow greenlet log warnings by removing the trace function from greenlet."""
    greenlet.settrace(None)



from .plugins import get_formatter_plugins, _true


def create_formatter(plugin_names):
    """
    Creates a formatter that uses a list of plugin names.
    :param plugin_names: list of plugin names
    :returns: formatter function
    """
    available_formatters = get_formatter_plugins(lambda k: k in plugin_names)
    formatters = [available_formatters[k]() for k in plugin_names]
    format_string = " ".join("{}={}".format(f.get_format_key(), f.get_format_string()) for f in formatters)
    return create_formatter_by_format_string(format_string, plugin_names)


def create_formatter_by_format_string(format_string, plugin_names):
    """
    Creates a formatter that uses a given format string and a list of plugin names.
    The format string needs to have fields in place to format all the plugins.
    :param format_string: format string (e.g. "a={} b={}")
    :param plugin_names: list of plugin names
    :returns: formatter function
    """
    formatter_classes = get_formatter_plugins(lambda k: k in plugin_names)
    formatters = [formatter_classes[k]() for k in plugin_names]
    getters = [f.get_value for f in formatters]

    def formatter(record, handler):
        return format_string.format(*(g(record) for g in getters))
    return formatter


def create_default_formatter(plugin_predicate=_true):
    """
    Creates a default formatter that uses a custom format string.

    The full format string is:
    <time> <hostname> pid=<pid>:<pname> tid=<tid>:<gid> tag=<rid_tag> module=<channel> level=<log_level> [other plugins] msg=<message>

    The builder takes into account plugins the are disabled to generate a more compact output, for example if procname
    if empty:
    pid=<pid>

    Or if gevent is not available/disabled:
    tid=<tid>

    :param plugin_predicate: predicate over plugin names to choose which plugins to use.
    :returns: formatter function
    """
    available_formatters = dict((k, v()) for k, v in get_formatter_plugins(plugin_predicate).items())

    strformats = []
    getters = []
    used_fields = set(['message'])  # we want message to be last

    def extend_format(str, *keys):
        used_fields.update(keys)
        getters.extend([available_formatters[f].get_value for f in keys])
        strformats.append(str.format(*[available_formatters[f].get_format_string() for f in keys]))

    if 'time' in available_formatters:
        extend_format('{}', 'time')

    if 'hostname' in available_formatters:
        extend_format('{}', 'hostname')

    if 'process_id' in available_formatters and 'procname' in available_formatters:
        extend_format('pid={}:{}', 'process_id', 'procname')
    elif 'process_id' in available_formatters:
        extend_format('pid={}', 'process_id')
    elif 'procname' in available_formatters:
        extend_format('pname={}', 'procname')

    if 'thread_id' in available_formatters and 'greenlet_id' in available_formatters:
        extend_format('tid={}:{}', 'thread_id', 'greenlet_id')
    elif 'thread_id' in available_formatters:
        extend_format('tid={}', 'thread_id')
    elif 'greenlet_id' in available_formatters:
        extend_format('gid={}', 'greenlet_id')

    if 'request_id_tag' in available_formatters:
        extend_format('tag={}', 'request_id_tag')
    if 'channel' in available_formatters:
        extend_format('module={}', 'channel')  # for backward compatibility we call this module
    if 'log_level' in available_formatters:
        extend_format('level={}', 'log_level')

    for f in sorted(set(available_formatters.keys()) - used_fields):
        extend_format(available_formatters[f].get_format_key() + '={}', f)

    if 'message' in available_formatters:
        extend_format('msg={}', 'message')

    strformat = " ".join(strformats)

    def formatter(record, handler):
        return strformat.format(*(g(record) for g in getters))
    return formatter



from infi.logging.plugins import InjectorPlugin, FormatterPlugin

try:
    import setproctitle

    def get_proc_title():
        return setproctitle.getproctitle()
except ImportError:
    def get_proc_title():
        return None


PROCNAME_KEY = 'procname'
_procname = None


def set_procname(s):
    """
    Sets the process name to be used in logging. This is useful if setproctitle package doesn't exist and you want to
    manually set the process name.
    :param s: process name
    :type s: str
    """
    global _procname
    _procname = s


def get_procname():
    """
    Returns the process name used for logging, calls `refresh_procname_if_none()` to make sure there's a process name.
    :returns: process name str (not None)
    """
    global _procname
    refresh_procname_if_none()
    return _procname


def refresh_procname_if_none(force=False):
    """
    If process name is None or `force` is True, fetches the process name by using `setproctitle` package and raises
    an error if it doesn't exist.
    :param force: even if there is already a process name, refresh it from `setproctitle`.
    """
    global _procname
    if _procname is None or force:
        _procname = get_proc_title()


def inject_procname(record):
    """
    Sets the process name on `record.extra`.
    :param record: logbook record
    """
    record.extra[PROCNAME_KEY] = get_procname()


def get_procname_from_record(record):
    """
    :param record: logbook record
    :returns: current process name retrieved from the logbook record
    """
    return record.extra.get(PROCNAME_KEY, '')


class ProcnameInjectorPlugin(InjectorPlugin):
    def inject(self, record):
        return inject_procname(record)


class ProcnameFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return get_procname_from_record(record)

    def get_format_string(self):
        return "{: <28}"

    def get_format_key(self):
        return "procname"



from infi.logging.plugins import FormatterPlugin


def get_log_level_name_from_record(record):
    """
    :param record: lobgook record
    :returns: record's log level name
    """
    return 'TRACE' if record.extra.get('TRACE', False) else record.level_name


class LogLevelFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return get_log_level_name_from_record(record)

    def get_format_string(self):
        return "{: <7}"

    def get_format_key(self):
        return "log_level"



from .rotating_file_handler import RotatingFileHandler

__all__ = ['RotatingFileHandler']
try:
    import infi.tracing
    from .syslog_handler import SyslogHandler
    __all__.append('SyslogHandler')
except ImportError:
    pass



import time as t
import logbook

_time = t.time
_logger = logbook.Logger('infi.logging')
_threadlocal_factory = None


def set_time_func(f):
    """
    Allow user to set the time function used for time measurements. By default the time function is Python's standard
    library time.time(), but the user can choose a different one, for example by using infi.monotonic_time.
    :param f: a function that has no arguments and returns a float representing seconds
    """
    global _time
    _time = f


def get_time_func():
    global _time
    return _time


def get_time():
    global _time
    return _time()


def set_logger(l):
    """
    Allow user to set the logger used to log the time measurements. By default the logger is a local logger created.
    :param l: logbook.Logger object
    """
    global _logger
    _logger = l


def get_logger():
    global _logger
    return _logger


def get_threadlocal_factory():
    global _threadlocal_factory
    if _threadlocal_factory is None:
        try:
            import gevent
            _threadlocal_factory = gevent.local.local
        except ImportError:
            import threading
            _threadlocal_factory = threading.local
    return _threadlocal_factory


def set_threadlocal_factory(factory):
    global _threadlocal_factory
    _threadlocal_factory = factory


def new_threadlocal(*args, **kwargs):
    return get_threadlocal_factory()(*args, **kwargs)



__version__ = "0.4.9"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = u'(HEAD detached at v0.4.9)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '401a5497ca6fadf8873c3884ee9d20f3f61199b7'
__git_head_subject__ = u'Finished Release v0.4.9'
__git_head_message__ = u''
__git_dirty_diff__ = u''
__git_commit_date__ = '2018-12-20 18:04:17'



from infi.logging.plugins import InjectorPlugin, FormatterPlugin
from socket import gethostname

HOST_NAME_KEY = 'host_name'


def get_host_name_from_record(record):
    """
    :param record: logbook record
    :returns: current thread ID from record
    """
    return record.extra.get(HOST_NAME_KEY, '')


def inject_host_name(record):
    """
    Sets the current thread ID on `record.extra`.
    :param record: logbook record
    """
    record.extra[HOST_NAME_KEY] = gethostname()


class HostNameInjectorPlugin(InjectorPlugin):
    def inject(self, record):
        return inject_host_name(record)


class HostNameFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return get_host_name_from_record(record)

    def get_format_string(self):
        return "{}"

    def get_format_key(self):
        return "hostname"



from infi.logging.plugins import FormatterPlugin


class ProcessIDFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return record.process

    def get_format_string(self):
        return "{:0>5}"

    def get_format_key(self):
        return "process_id"



import logbook
from infi.logging.plugins import get_injector_plugins, _true


def create_inject_extra_data(plugin_predicate=_true):
    """
    Composes all the injector plugins and returns a data injection function.
    :param plugin_predicate: predicate that chooses which plugins to enable
    :returns: function(record: logbook record)
    """
    injectors = [injector() for injector in get_injector_plugins(plugin_predicate).values()]

    def inject_extra_data(record):
        """Wrapper that injects all the data available from the loaded injector plugins"""
        for injector in injectors:
            injector.inject(record)
    return inject_extra_data


def create_processor(plugin_predicate=_true):
    """
    Returns a new logbook Processor object that injects to each record's `extra` dict whatever the enabled plugins
    choose to put there.
    :param plugin_predicate: predicate that chooses which plugins to enable
    :returns: logbook.Processor object
    """
    return logbook.Processor(create_inject_extra_data(plugin_predicate))



from infi.logging.plugins import FormatterPlugin


def get_channel_from_record(record):
    """
    :param record: lobgook record
    :returns: record's log module
    """
    return record.channel


class ChannelFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return get_channel_from_record(record)

    def get_format_string(self):
        return "{: <40}"

    def get_format_key(self):
        return "channel"




SETUP_INFO = dict(
    name = 'infi.logging',
    version = '0.4.9',
    author = 'Arnon Yaari',
    author_email = 'arnony@infinidat.com',

    url = 'https://github.com/Infinidat/infi.logging',
    license = 'BSD',
    description = """Extensions for logbook, such as Windows EventLog handler.""",
    long_description = """Extensions for logbook, such as Windows EventLog handler.""",

    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],

    install_requires = [
'infi.registry',
'Logbook',
'setuptools',
'six'
],
    namespace_packages = ['infi'],

    package_dir = {'': 'src'},
    package_data = {'': ['*dll']},
    include_package_data = True,
    zip_safe = False,

    entry_points = dict(
        console_scripts = [],
        gui_scripts = [],
        infi_logging_formatter_plugins = [
'channel = infi.logging.plugins.channel:ChannelFormatterPlugin',
'log_level = infi.logging.plugins.log_level:LogLevelFormatterPlugin',
'message = infi.logging.plugins.message:MessageFormatterPlugin',
'hostname = infi.logging.plugins.hostname:HostNameFormatterPlugin',
'greenlet_id = infi.logging.plugins.greenlet_id:GreenletIDFormatterPlugin',
'procname = infi.logging.plugins.procname:ProcnameFormatterPlugin',
'request_id_tag = infi.logging.plugins.request_id_tag:RequestIDTagFormatterPlugin',
'process_id = infi.logging.plugins.process_id:ProcessIDFormatterPlugin',
'thread_id = infi.logging.plugins.thread_id:ThreadIDFormatterPlugin',
'time = infi.logging.plugins.time:TimeFormatterPlugin'],
        infi_logging_injector_plugins = [
'hostname = infi.logging.plugins.hostname:HostNameInjectorPlugin',
'greenlet_id = infi.logging.plugins.greenlet_id:GreenletIDInjectorPlugin',
'procname = infi.logging.plugins.procname:ProcnameInjectorPlugin',
'request_id_tag = infi.logging.plugins.request_id_tag:RequestIDTagInjectorPlugin',
'thread_id = infi.logging.plugins.thread_id:ThreadIDInjectorPlugin']
        ),
)

if SETUP_INFO['url'] is None:
    _ = SETUP_INFO.pop('url')

def setup():
    from setuptools import setup as _setup
    from setuptools import find_packages
    SETUP_INFO['packages'] = find_packages('src')
    _setup(**SETUP_INFO)

if __name__ == '__main__':
    setup()




from infi.logging.plugins import InjectorPlugin, FormatterPlugin
from gevent.thread import get_ident

GREENLET_ID_KEY = 'greenlet_id'
GREENLET_ID_MODULU = 32768


def get_greenlet_id_from_record(record):
    """
    :param record: lobgook record
    :returns: greenlet ID of the logbook record or -1 if not set
    """
    return record.extra.get(GREENLET_ID_KEY, -1)


def inject_greenlet_id(record):
    """
    Sets the greenlet ID on `record.extra`.
    :param record: logbook record
    """
    record.extra[GREENLET_ID_KEY] = get_ident()


class GreenletIDInjectorPlugin(InjectorPlugin):
    def inject(self, record):
        return inject_greenlet_id(record)


class GreenletIDFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return get_greenlet_id_from_record(record) % GREENLET_ID_MODULU

    def get_format_string(self):
        return "{:0>5}"

    def get_format_key(self):
        return "greenlet_id"



import logbook
from logbook.handlers import rename, sys, errno


class RotatingFileHandler(logbook.RotatingFileHandler):
    """Similar to logbook's `RotatingFileHandler`, but the backup files use leading zeros in the suffix so it'll be
    easier to sort when listing files (e.g. x.01, x.02 and not x.1, x.2)."""
    def perform_rollover(self):
        # Code here is almost the same as the base class's perform_rollover except for the src and dst formatting.
        self.stream.close()
        for x in range(self.backup_count - 1, 0, -1):
            src = '{}.{:02}'.format(self._filename, x)
            dst = '{}.{:02}'.format(self._filename, x + 1)
            try:
                rename(src, dst)
            except OSError:
                e = sys.exc_info()[1]
                if e.errno != errno.ENOENT:
                    raise
        rename(self._filename, self._filename + '.01')
        self._open('w')



"""
Greenlet/thread-friendly request ID tagging for log messages.
"""
import logbook
import random
from infi.pyutils.decorators import wraps
from infi.pyutils.contexts import contextmanager

from infi.logging.plugins import InjectorPlugin, FormatterPlugin
from infi.logging.globals import new_threadlocal, get_logger


REQUEST_ID_TAG_KEY = 'request_id'
_threadlocal = None


def _get_threadlocal():
    global _threadlocal
    if _threadlocal is None:  # Lazy creation of threadlocal so user can configure which threadlocal to use
        _threadlocal = new_threadlocal()
    return _threadlocal


def get_tag():
    """
    :returns: current request ID tag for the current greenlet if exists or None if no request ID tag is set
    :rtype: str or None
    """
    return getattr(_get_threadlocal(), REQUEST_ID_TAG_KEY, None)


def set_tag(tag):
    """
    Sets the request ID tag on the current greenlet.
    :param tag: tag string for the current request
    :type tag: str or None to clear the tag
    """
    setattr(_get_threadlocal(), REQUEST_ID_TAG_KEY, tag)


def new_random_tag():
    """
    :returns: new random str (8 character random hex string)
    :rtype: str
    """
    return hex(random.getrandbits(32))[2:-1].zfill(8)


def set_random_tag():
    """
    Sets a new random tag on the current greenlet.
    """
    set_tag(new_random_tag())


def get_request_id_tag_from_record(record):
    """
    :returns: request ID tag set on the logbook record or None if not found
    :rtype: str or None
    """
    return record.extra.get(REQUEST_ID_TAG_KEY, '')


def inject_request_id_tag(record):
    """Logbook processor function that sets REQUEST_ID_TAG_KEY attribute on the log record to the current tag (or None)"""
    tag = get_tag()
    if tag is not None:
        record.extra[REQUEST_ID_TAG_KEY] = tag


@contextmanager
def _null_context():
    yield


@contextmanager
def request_id_tag_context(title=None, tag=None, logger=None):
    """
    Context that adds a request ID tag logging processor.

    :param title: title to write in the log when setting a new tag (if a previous tag was not set).
                  If ``None`` no log message will be generated.
    :param tag: tag to set before calling func. If tag is None and no existing tag is set a new random tag will be
                created.
    :param logger: logger to use, if None use default logger
    """
    if logger is None:
        logger = get_logger()

    prev_tag, new_tag = get_tag(), tag
    if new_tag is not None:
        set_tag(new_tag)
    elif prev_tag is None:
        new_tag = new_random_tag()
        set_tag(new_tag)

    # We create a logbook.Processor context only if we didn't have a previous tag, otherwise there must already
    # be a context in place somewhere down the call stack.
    with (logbook.Processor(inject_request_id_tag).greenletbound() if prev_tag is None else _null_context()):
        if prev_tag is None and title is not None:
            # Log this function since it's our first "tagged" entry to the greenlet
            logger.debug("setting new tag {} on greenlet {}".format(new_tag, title))
        try:
            yield
        finally:
            set_tag(prev_tag)


def request_id_tag(func=None, tag=None, logger=None):
    """
    Decorator that wraps func and adds a request ID tag logging processor that adds the request ID attribute to log
    records within the same greenlet.

    It can be used in several methods:
    ```
    @request_id_tag  # this will use existing tag on the greenlet or a new random tag
    def foo():
        some_logger.info("hi ya!")

    @request_id_tag(tag='hardcode')  # this will set the tag to 'hardcore' from hereon on this greenlet
    def foo():
        some_logger.info("hi ya!")

    wrapped_func = request_id_tag(foo, tag=get_tag())  # call explicitly so we can pass tag between greenlets
    gevent.spawn(wrapped_func)
    ```

    :param func: function to wrap
    :param tag: tag to set before calling func. If tag is None and no existing tag is set a new random tag will be
                created.
    :param logger: logger to use, if None use default logger
    :returns: wrapped function
    """
    def decorate(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            with request_id_tag_context(f.__name__, tag, logger):
                return f(*args, **kwargs)
        return wrapped

    if func is None:
        return decorate
    else:
        return decorate(func)


class RequestIDTagInjectorPlugin(InjectorPlugin):
    def inject(self, record):
        return inject_request_id_tag(record)


class RequestIDTagFormatterPlugin(FormatterPlugin):
    def get_value(self, record):
        return get_request_id_tag_from_record(record)

    def get_format_string(self):
        return "{:0>8}"

    def get_format_key(self):
        return "request_id_tag"
